import { Component, ViewChild, ElementRef } from "@angular/core";

import { PabResults } from "../../results/pab-results";

import * as XLSX from "xlsx";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { I18nService } from "../../services/internationalisation/internationalisation.service";
import { ResultsComponent } from "../fixedvar-results/results.component";

@Component({
    selector: "pab-results-table",
    templateUrl: "./pab-results-table.component.html",
    styleUrls: [
        "./pab-results-table.component.scss"
    ]
})
export class PabResultsTableComponent extends ResultsComponent {

    /** résultats non mis en forme */
    private _pabResults: PabResults;

    /** entêtes des colonnes */
    private _headers: string[];

    /** résultats mis en forme */
    private _dataSet: any[];

    @ViewChild("tableContainer") table: ElementRef;

    constructor(
        protected appSetupService: ApplicationSetupService,
        protected intlService: I18nService
    ) {
        super();
    }

    public set results(r: PabResults) {
        this._pabResults = r;

        this._dataSet = [];
        if (
            this._pabResults
            && this._pabResults.cloisonsResults
            && this._pabResults.cloisonsResults.length > 0
            && ! this._pabResults.hasError()
        ) {
            const pr = this._pabResults;
            const nDigits = this.appSetupService.displayDigits;

            // refresh headers here if language changed
            this._headers = pr.headers;

            // line 1
            if (pr.cloisonsResults[0].vCalc) { // parfois le calcul des cloisons échoue
                this._dataSet.push([
                    "",
                    pr.cloisonsResults[0] ? pr.cloisonsResults[0].vCalc.toFixed(nDigits) : "",
                    "", "", "", "", "", "", ""
                ]);
            }

            // lines 2 - n-1
            for (let i = 0; i < pr.cloisonsResults.length - 1; i++) {
                if (
                    pr.cloisonsResults[i].vCalc
                    && pr.cloisonsResults[i + 1].vCalc
                ) {
                    const ZRAM = pr.cloisonsResults[i].getExtraResult("ZRAM");
                    const PV = pr.cloisonsResults[i].getExtraResult("PV");
                    const YMOY = pr.cloisonsResults[i].getExtraResult("YMOY");
                    const ZRB = pr.cloisonsResults[i].getExtraResult("ZRB");
                    this._dataSet.push([
                        i + 1, // n° cloison
                        pr.cloisonsResults[i + 1].vCalc.toFixed(nDigits), // Z
                        (ZRAM !== undefined) ? ZRAM.toFixed(nDigits) : "", // ZRAM
                        (pr.cloisonsResults[i].vCalc - pr.cloisonsResults[i + 1].vCalc).toFixed(nDigits), // DH
                        pr.cloisonsQ[i].toFixed(nDigits), // Q
                        (PV !== undefined) ? PV.toFixed(nDigits) : "", // PV
                        (YMOY !== undefined) ? YMOY.toFixed(nDigits) : "", // YMOY
                        (ZRB !== undefined) ? ZRB.toFixed(nDigits) : "", // ZRB
                        pr.bassinsQA[i] // QA
                    ]);
                }
            }

            // line n
            const l = pr.cloisonsResults.length;
            if (
                pr.cloisonsResults[l - 1].vCalc
                && pr.cloisonAvalResults.vCalc
            ) {
                const nZRAM = pr.cloisonsResults[l - 1].getExtraResult("ZRAM");
                const nPV = pr.cloisonsResults[l - 1].getExtraResult("PV");
                const nY = pr.cloisonsResults[l - 1].getExtraResult("YMOY");
                const nZRB = pr.cloisonsResults[l - 1].getExtraResult("ZRB");
                this._dataSet.push([
                    l, // n° cloison
                    pr.cloisonAvalResults.vCalc.toFixed(nDigits), // Z
                    (nZRAM !== undefined) ? nZRAM.toFixed(nDigits) : "", // ZRAM
                    (pr.cloisonsResults[l - 1].vCalc - pr.cloisonAvalResults.vCalc).toFixed(nDigits), // DH
                    pr.cloisonsQ[l - 1].toFixed(nDigits), // Q
                    (nPV !== undefined) ? nPV.toFixed(nDigits) : "", // PV
                    (nY !== undefined) ? nY.toFixed(nDigits) : "", // YMOY
                    (nZRB !== undefined) ? nZRB.toFixed(nDigits) : "", // ZRB
                    pr.bassinsQA[l - 1] // QA
                ]);
            }

            // downstream line
            if (pr.cloisonAvalResults.vCalc) {
                const caZRAM = pr.cloisonAvalResults.getExtraResult("ZRAM");
                this._dataSet.push([
                    this.intlService.localizeText("INFO_LIB_AVAL"), // n° cloison
                    pr.Z2.toFixed(nDigits), // Z
                    (caZRAM !== undefined) ? caZRAM.toFixed(nDigits) : "", // ZRAM
                    (pr.cloisonAvalResults.vCalc - pr.Z2).toFixed(nDigits), // DH
                    pr.cloisonAvalQ.toFixed(nDigits), // Q
                    "", "", "", ""
                ]);
            }
        }
    }

    public get hasResults(): boolean {
        return this._pabResults && this._pabResults.hasResults;
    }

    public get headers() {
        return this._headers;
    }

    /**
     * Returns a combination of and results and extraResults for mat-table
     */
    public get dataSet() {
        return this._dataSet;
    }

    public exportAsSpreadsheet() {
        // automagic <table> extraction
        const ws: XLSX.WorkSheet = XLSX.utils.table_to_sheet(this.table.nativeElement);
        const wb: XLSX.WorkBook = XLSX.utils.book_new();
        XLSX.utils.book_append_sheet(wb, ws, "default");
        // save and download
        XLSX.writeFile(wb, "PABResults.xlsx");
    }
}
