import { Component, Input, Output, EventEmitter, OnInit } from "@angular/core";

import { MatDialog } from "@angular/material";

import {
    Pab,
    Session,
    PabCloisons,
    Props,
    CalculatorType,
    ParallelStructure,
    LoiDebit,
    Cloisons,
    Nub,
    Structure
 } from "jalhyd";

import { I18nService } from "../../services/internationalisation/internationalisation.service";
import { FormulaireService } from "../../services/formulaire/formulaire.service";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { PabTable } from "../../formulaire/pab-table";
import { DialogEditPabComponent } from "../dialog-edit-pab/dialog-edit-pab.component";

/**
 * The big editable data grid for calculator type "Pab" (component)
 */
@Component({
    selector: "pab-table",
    templateUrl: "./pab-table.component.html",
    styleUrls: [
        "./pab-table.component.scss"
    ]
})
export class PabTableComponent implements /* DoCheck, AfterViewInit, */ OnInit {

    @Input()
    private pabTable: PabTable;

    /** flag de validité des FieldSet enfants */
    private _isValid = false;

    /** événément de changement de validité */
    @Output()
    private validChange = new EventEmitter();

    /** événément de changement de valeur d'un input */
    @Output()
    private inputChange = new EventEmitter();

    /** underlying Pab, binded to the rows */
    private model: Pab;

    /** general headers above the columns */
    public headers: any[];

    /** columns headers description */
    public cols: any[];

    /** data binded to the table */
    public rows: any[];

    /** number of children to add when clicking "add" or "clone" button */
    public childrenToAdd = 1;

    /** items currently selected */
    private selectedItems: Nub[];

    /** used for shift+click implementation */
    private latestClickedCell: any;

    public constructor(
        private i18nService: I18nService,
        private formService: FormulaireService,
        private editPabDialog: MatDialog,
        private appSetupService: ApplicationSetupService
    ) {
        this.selectedItems = [];
    }

    public get title(): string {
        return this.i18nService.localizeText("INFO_PAB_TABLE");
    }

    public get isValid() {
        return this._isValid;
    }

    /** returns true if the cell has an underlying model (ie. is editable) */
    public hasModel(cell: any): boolean {
        return (cell !== undefined && cell.model !== undefined);
    }

    /** returns true if the cell is an editable number */
    public isNumberInput(cell: any): boolean {
        return this.hasModel(cell) && ! this.isSelect(cell);
    }

    /** returns true if the cell is a select box @TODO rename */
    public isSelect(cell: any): boolean {
        return this.hasModel(cell) && (cell.options !== undefined);
    }

    /** value to display in a cell, if it is not editable */
    public cellValue(cell: any) {
        if (cell === undefined) {
            return "";
        } else {
            if (this.hasModel(cell)) {
                return cell.model;
            } else {
                return cell.value;
            }
        }
    }

    public rowSpan(cell: any) {
        if (cell !== undefined && cell.rowspan) {
            return cell.rowspan;
        }
        return undefined;
    }

    public colSpan(cell: any) {
        if (cell !== undefined && cell.colspan) {
            return cell.colspan;
        }
        return undefined;
    }


    /** returns true if the cell / row has a selectable item */
    public isSelectable(cellOrRow: any): boolean {
        return (
            cellOrRow !== undefined
            && cellOrRow.selectable
        );
    }

    /** returns true if the cell / row has a selectableColumn item */
    public isSelectableByColumn(cellOrRow: any): boolean {
        return (
            cellOrRow !== undefined
            && cellOrRow.selectableColumn
        );
    }

    /**
     *  - checks if the cell / row has a selectable item, that is currently
     *      selected
     *  - if cell / row has a selectableColumn attribute, also checks if
     *      this column is selected
     *
     *  returns true if at least one criterion is met
     */
    public isSelected(cellOrRow: any): boolean {
        let cellSelected = false;
        let columnSelected = false;
        // cell
        if (this.isSelectable(cellOrRow)) {
            cellSelected = true;
            if (Array.isArray(cellOrRow.selectable)) {
                for (const elt of cellOrRow.selectable) {
                    cellSelected = cellSelected && this.selectedItems.includes(elt);
                }
            } else {
                cellSelected = this.selectedItems.includes(cellOrRow.selectable);
            }
        }
        // column
        if (this.isSelectableByColumn(cellOrRow)) {
            columnSelected = this.isDeviceColumnSelected(cellOrRow.selectableColumn);
        }
        // done
        return (cellSelected || columnSelected);
    }

    /**
     * returns true if every wall (including downwall) has its nth device
     * selected (or has no nth device)
     */
    public isDeviceColumnSelected(n: number): boolean {
        let ok = true;
        for (const c of this.model.children) {
            const nthChild = c.getChildren()[n];
            if (nthChild) {
                ok = ok && this.selectedItems.includes(nthChild);
            }
        }
        const nthChildDW = this.model.downWall.getChildren()[n];
        if (nthChildDW) {
            ok = ok && this.selectedItems.includes(nthChildDW);
        }
        return ok;
    }

    /**
     * selects or unselects the clicked cell, depending on its current state
     * and the modifier key held if any
     */
    public toggleSelection(cell: any, $event: any) {
        if (
            this.isSelectable(cell)
            && ! this.hasModel(cell) // editable cells listen to the click event for edition only
        ) {
            if ($event.shiftKey && cell !== this.latestClickedCell) { // shift + click
                // @TODO interpopolate from this.latestClickedCell to this one
                console.log("shift + click");

            } else if (
                $event.ctrlKey // ctrl + click
                || ($event.shiftKey && cell === this.latestClickedCell) // shift on same cell => equiv. of ctrl
            ) {
                if (this.isSelected(cell)) {
                    // unselect this cell / these cells
                    if (Array.isArray(cell.selectable)) {
                        this.selectedItems = this.selectedItems.filter(e => ! cell.selectable.includes(e));
                    } else {
                        this.selectedItems = this.selectedItems.filter(e => e !== cell.selectable);
                    }
                } else {
                    // add this cell / these cells to selection
                    if (Array.isArray(cell.selectable)) {
                        this.selectedItems = this.selectedItems.concat(cell.selectable);
                        this.selectedItems = this.selectedItems.filter(
                            (item, index) => this.selectedItems.indexOf(item) === index // deduplicate
                        );
                    } else {
                        this.selectedItems.push(cell.selectable);
                    }
                }

            } else { // just a click
                if (this.isSelected(cell)) {
                    // select nothing
                    this.selectedItems = [];
                } else {
                    // select this cell / thses cells only
                    if (Array.isArray(cell.selectable)) {
                        this.selectedItems = cell.selectable.slice(); // array copy
                    } else {
                        this.selectedItems = [ cell.selectable ];
                    }
                }
            }
            this.latestClickedCell = cell;

            // clean list
            this.selectedItems = this.selectedItems.filter(e => e !== undefined);

            // @TODO useful ?
            $event.preventDefault();
            $event.stopPropagation();
            return false;
        }
    }

    // quick getter for 1st selected item
    public get selectedItem() {
        if (this.selectedItems.length === 0) {
            throw new Error("get selectedItem() : no item selected");
        }
        return this.selectedItems[0];
    }

    // prevents Firefox to display weird cell border when ctrl+clicking
    public preventCtrlClickBorder($event) {
        if ($event.ctrlKey) {
            $event.preventDefault();
        }
    }

    public get addManyOptionsList() {
        return Array(20).fill(0).map((value, index) => index + 1);
    }

    // DEBUG
    private dumpParams(n: Nub) {
        console.log(`---- PARAMETERS FOR NUB "${n.uid}" (${n.constructor.name}) ----`);
        for (const p of n.parameterIterator) {
            if (p.visible) {
                console.log(`> ${p.symbol} : ${p.singleValue}, ${p.currentValue}, ${p.v}`);
            }
        }
    }

    public zpouet() {
        console.log("!!!------------ zpoueeeeeeeeet ------------!!!");
        for (const c of this.model.children) {
            this.dumpParams(c);
        }
        this.dumpParams(this.model.downWall);
    }

    // at this time @Input data is supposed to be already populated
    public ngOnInit() {
        // get model
        this.model = this.pabTable.pab;

        // ------------- TEST ------------------------

        // empty PAB
        const chLen = this.model.children.length;
        for (let i = 0; i < chLen; i++) {
            this.model.deleteChild(0);
        }

        // cloison à 1 ouvrage
        const cl1 = Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Cloisons })
        ) as Cloisons;
        cl1.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.RectangularOrificeFree }), cl1
        ));
        // PabCloison associé
        const pc1 = Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.PabCloisons }), this.model
        ) as PabCloisons;
        pc1.initModelCloisons(cl1);
        pc1.prms.setCurrentValuesFromModel();
        this.model.addChild(pc1);
        this.dumpParams(pc1);

        // cloison à 3 ouvrages
        const cl2 = Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Cloisons })
        ) as Cloisons;
        cl2.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.GateCem88v }), cl2
        ));
        cl2.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.RectangularOrificeFree }), cl2
        ));
        cl2.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.RectangularOrificeSubmerged }), cl2
        ));
        // PabCloison associé
        const pc2 = Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.PabCloisons }), this.model
        ) as PabCloisons;
        pc2.initModelCloisons(cl2);
        pc2.prms.setCurrentValuesFromModel();
        this.model.addChild(pc2);
        this.dumpParams(pc2);

        // cloison à 2 ouvrages
        const cl3 = Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Cloisons })
        ) as Cloisons;
        cl3.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.GateCem88v }), cl3
        ));
        cl3.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.RectangularOrificeFree }), cl3
        ));
        // PabCloison associé
        const pc3 = Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.PabCloisons }), this.model
        ) as PabCloisons;
        pc3.initModelCloisons(cl3);
        pc3.prms.setCurrentValuesFromModel();
        this.model.addChild(pc3);
        this.dumpParams(pc3);

        // cloison aval à 1 ouvrage
        const cloisonAval: ParallelStructure = Session.getInstance().createSessionNub(
            new Props({ calcType: CalculatorType.ParallelStructure })
        ) as ParallelStructure;
        cloisonAval.addChild(Session.getInstance().createNub(
            new Props({ calcType: CalculatorType.Structure, loiDebit: LoiDebit.KIVI }), cloisonAval
        ));
        this.model.setDownWall(cloisonAval.uid);
        this.dumpParams(this.model.downWall);

        // debug
        console.log("Model in ngOnInit:", this.model);

        // ------------- FIN TEST --------------------

        this.refresh();
    }

    /**
     * Builds the editable data grid from the Pab model
     */
    private refresh() {
        const maxNbDevices = this.findMaxNumberOfDevices();

        // 0. build spanned headers over real columns
        this.headers = [];
        // 1 header for basin
        let bs: any[] = this.model.children;
        bs = bs.concat(this.model.downWall);
        this.headers.push({
            title: "Bassin",
            colspan: 6,
            selectable: bs
        });
        // 1 header for each device of the wall having the most devices (including downwall)
        for (let i = 0; i < maxNbDevices; i++) {
            this.headers.push({
                title: "Cloison : ouvrage n°" + (i + 1),
                colspan: 3,
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
        }

        // A. build columns set
        this.cols = [];
        // 6 cols for basin
        this.cols.push({
            title: "N° de bassin",
            selectable: bs
        });
        this.cols.push({
            title: "Longueur",
            selectable: bs
        });
        this.cols.push({
            title: "Largeur",
            selectable: bs
        });
        this.cols.push({
            title: "Débit d'attrait",
            selectable: bs
        });
        this.cols.push({
            title: "Cote radier mi-bassin",
            selectable: bs
        });
        this.cols.push({
            title: "Cote radier amont paroi",
            selectable: bs
        });
        // no col for wall type (defined by rowspan-2 header above)
        // 3 cols for each device of the wall having the most devices (including downwall)
        for (let i = 0; i < maxNbDevices; i++) {
            this.cols.push({
                title: "Type",
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
            this.cols.push({
                title: "Paramètres",
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
            this.cols.push({
                title: "Valeurs",
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
        }

        // B. Build rows set
        this.rows = [];
        // B.1 many rows for each wall
        let childIndex = 0;
        for (const cloison of this.model.children) {
            // as much rows as the greatest number of parameters among its devices
            const maxNbParams = this.findMaxNumberOfDeviceParameters(cloison);
            // console.log(">>> max nb params: ", maxNbParams);
            for (let i = 0; i < maxNbParams; i++) {
                // build device params row
                const deviceParamRow = { selectable: cloison, cells: [] };
                // basin number
                if (i === 0) {
                    deviceParamRow.cells.push({
                        value: childIndex + 1,
                        rowspan: maxNbParams + 1,
                        class: "basin_number",
                        selectable: cloison
                    });
                }
                // 5 empty cells
                if (i === 0) {
                    deviceParamRow.cells.push({
                        colspan: 5,
                        rowspan: maxNbParams,
                        selectable: cloison
                    });
                }
                // device param cells : 3 cells for each device
                for (const ouvrage of cloison.structures) {
                    const nvParam = ouvrage.getNthVisibleParam(i);
                    // cell 1 : device type
                    if (i === 0) { // 1st row
                        deviceParamRow.cells.push({
                            model: ouvrage.properties.getPropValue("loiDebit"),
                            // options: StructureProperties.findCompatibleLoiDebit(ouvrage)
                            options: [ "salut", "coucou", "pouet" ],
                            selectable: ouvrage
                        });
                    }
                    // fill space
                    if (i === 1) {
                        deviceParamRow.cells.push({
                            rowspan: (maxNbParams - 1),
                            selectable: ouvrage
                        });
                    }
                    // cell 2 : param name
                    if (nvParam) {
                        deviceParamRow.cells.push({
                            value: nvParam.symbol,
                            selectable: ouvrage
                        });
                    } else {
                        deviceParamRow.cells.push({
                            selectable: ouvrage
                        });
                    }
                    // cell 3 : param value
                    if (nvParam) {
                        deviceParamRow.cells.push({
                            model: nvParam,
                            selectable: ouvrage
                        });
                    } else {
                        deviceParamRow.cells.push({
                            selectable: ouvrage
                        });
                    }
                }
                // fill horizontal space
                const devDiff =  (maxNbDevices - cloison.structures.length);
                if (i === 0) {
                    for (let j = 0; j < devDiff; j++) {
                        deviceParamRow.cells.push({
                            colspan: 3,
                            rowspan: maxNbParams,
                            selectable: cloison,
                            selectableColumn: cloison.structures.length + j
                        });
                    }
                }
                // done !
                this.rows.push(deviceParamRow);
            }
            // 1 row for the basin after the wall
            const basinRow: { selectable: any, cells: any[] } = {
                selectable: cloison,
                cells: [
                    // no cell for basin number (defined by rowspan-n cell above)
                    {
                        model: cloison.prms.LB
                    },
                    {
                        model: cloison.prms.BB
                    },
                    {
                        model: cloison.prms.QA
                    },
                    {
                        model: cloison.prms.ZRMB
                    },
                    {
                        model: cloison.prms.ZRAM
                    }
                ]
            };
            // fill horizontal space
            for (let i = 0; i < maxNbDevices; i++) {
                basinRow.cells.push({
                    colspan: 3
                });
            }
            // done !
            this.rows.push(basinRow);
            childIndex ++;
        }

        // B.2 many rows for downwall
        // as much rows as the greatest number of parameters among its devices
        const maxNbParamsDW = this.findMaxNumberOfDeviceParameters(this.model.downWall);
        for (let i = 0; i < maxNbParamsDW; i++) {
            // build device params row
            const deviceParamRowDW = { selectable: this.model.downWall, cells: [] };
            // basin number
            if (i === 0) {
                deviceParamRowDW.cells.push({
                    value: "Aval",
                    rowspan: maxNbParamsDW,
                    class: "basin_number",
                    selectable: this.model.downWall
                });
            }
            // 5 empty cells
            if (i === 0) {
                deviceParamRowDW.cells.push({
                    colspan: 5,
                    rowspan: maxNbParamsDW,
                    selectable: this.model.downWall
                });
            }
            // downwall device param cells : 3 cells for each device
            for (const ouvrage of this.model.downWall.structures) {
                const nvParam = ouvrage.getNthVisibleParam(i);
                // cell 1 : device type
                if (i === 0) { // 1st row
                    deviceParamRowDW.cells.push({
                        model: ouvrage.properties.getPropValue("loiDebit"),
                        // options: StructureProperties.findCompatibleLoiDebit(ouvrage)
                        options: [ "salut", "coucou", "pouet" ]
                    });
                }
                // fill space
                if (i === 1) {
                    deviceParamRowDW.cells.push({
                        rowspan: (maxNbParamsDW - 1),
                        selectable: ouvrage
                    });
                }
                // cell 2 : param name
                if (nvParam) {
                    deviceParamRowDW.cells.push({
                        value: nvParam.symbol,
                        selectable: ouvrage
                    });
                } else {
                    deviceParamRowDW.cells.push({
                        selectable: ouvrage
                    });
                }
                // cell 3 : param value
                if (nvParam) {
                    deviceParamRowDW.cells.push({
                        model: nvParam,
                        selectable: ouvrage
                    });
                } else {
                    deviceParamRowDW.cells.push({
                        selectable: ouvrage
                    });
                }
            }
            // fill horizontal space
            const devDiff =  (maxNbDevices - this.model.downWall.structures.length);
            if (i === 0) {
                for (let j = 0; j < devDiff; j++) {
                    deviceParamRowDW.cells.push({
                        colspan: 3,
                        rowspan: maxNbParamsDW,
                        selectable: this.model.downWall,
                        selectableColumn: this.model.downWall.structures.length + j
                    });
                }
            }
            // done !
            this.rows.push(deviceParamRowDW);
        }
    }

    private findMaxNumberOfDevices(): number {
        let maxNbDevices = 1;
        for (const w of this.model.children) {
            maxNbDevices = Math.max(maxNbDevices, w.getChildren().length);
        }
        maxNbDevices = Math.max(maxNbDevices, this.model.downWall.getChildren().length);
        return maxNbDevices;
    }

    private findMaxNumberOfDeviceParameters(struct: ParallelStructure): number {
        let maxNbParams = 1;
        for (const d of struct.getChildren()) {
            let nbParams = 0;
            for (const p of d.parameterIterator) {
                if (p.visible) {
                    // console.log("(counting)", p.symbol);
                    nbParams ++;
                }
            }
            // console.log(">>> child params: ", nbParams);
            maxNbParams = Math.max(maxNbParams, nbParams);
        }
        return maxNbParams;
    }

    /** returns true if exactly one device is selected, and nothing else */
    public get selectionIsOneDevice() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem instanceof Structure
        );
    }

    /**
     * Returns true if there is at least one selected item,
     * and all selected items are devices
     */
    private onlyDevicesAreSelected() {
        let ok = false;
        if (this.selectedItems.length > 0) {
            ok = true;
            for (const s of this.selectedItems) {
                ok = ok && (s instanceof Structure);
            }
        }
        return ok;
    }

    /**
     * Returns true if there is at least one selected item,
     * and all selected items are walls
     */
    private onlyWallsAreSelected(excludeDownwall: boolean = true) {
        let ok = false;
        if (this.selectedItems.length > 0) {
            ok = true;
            for (const s of this.selectedItems) {
                if (excludeDownwall) {
                    ok = ok && (s instanceof Cloisons);
                } else {
                    ok = ok && (s instanceof ParallelStructure);
                }
            }
        }
        return ok;
    }

    public get relatedEntityTitle() {
        let title = "";
        if (this.onlyDevicesAreSelected()) {
            title = "Ouvrages";
        } else if (this.onlyWallsAreSelected()) {
            title = "Cloisons";
        }
        if (title !== "") {
            title += " :";
        }
        return title;
    }

    public get enableAddButton() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem.parent // exclude downwall
        );
    }

    public get enableCopyButton() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem.parent // exclude downwall
        );
    }

    public get enableUpButton() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem.parent
            && this.selectedItem.findPositionInParent() !== 0
        );
    }

    public get enableDownButton() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem.parent
            && this.selectedItem.findPositionInParent() < (this.selectedItem.parent.getChildren().length - 1)
        );
    }

    public get enableRemoveButton() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem.parent
            && this.selectedItem.parent.getChildren().length > 1
        );
    }

    /**
     * returns true if at least one object is selected
     */
    public get enableEditPabButton() {
        return this.selectedItems.length > 0;
    }

    public onAddClick() {
        console.log("Add !!!");
        /* const newChild = undefined;
        this.selectedItem.parent.addChild(newChild, this.selectedItem.findPositionInParent()); */
        /* for (let i = 0; i < this.childrenToAdd; i++) {
            this.selectedItem.parent.addChild()
        } */
        /* let msg: string;
        if (this.childrenToAdd === 1) {
            msg = this.i18nService.localizeText("INFO_FSC_FS_ADDED");
        } else {
            msg = sprintf(this.i18nService.localizeText("INFO_FSC_FS_ADDED_N_TIMES"), this.childrenToAdd);
        }
        this.notifService.notify(msg); */
        this.childrenToAdd = 1; // reinit to avoid confusion
    }

    public onCopyClick() {
        for (let i = 0; i < this.childrenToAdd; i++) {
            const newChild = Session.getInstance().createNub(
                this.selectedItem.properties.clone(),
                this.selectedItem.parent
            );
            // copy parameter values
            for (const p of this.selectedItem.prms) {
                newChild.getParameter(p.symbol).singleValue = p.singleValue;
            }
            // copy children
            if (this.selectedItem instanceof ParallelStructure) {
                for (const c of this.selectedItem.getChildren()) {
                    const newGrandChild = Session.getInstance().createNub(
                        c.properties.clone(),
                        newChild
                    );
                    // copy children parameters values
                    for (const p of c.prms) {
                        newGrandChild.getParameter(p.symbol).singleValue = p.singleValue;
                    }
                    // add to parent
                    newChild.addChild(
                        newGrandChild,
                        c.findPositionInParent()
                    );
                }
            }
            // add to parent
            this.selectedItem.parent.addChild(
                newChild,
                this.selectedItem.findPositionInParent()
            );
        }
        this.refresh();
        /* const pos = (this._fieldSet.parent as FieldsetContainer).getFieldsetPosition(this._fieldSet) + 1;
        let msg: string;
        if (this.childrenToAdd === 1) {
            msg = sprintf(this.i18nService.localizeText("INFO_FSC_FS_COPIED"), pos);
        } else {
            msg = sprintf(this.i18nService.localizeText("INFO_FSC_FS_COPIED_N_TIMES"), pos, this.childrenToAdd);
        }
        this.notifService.notify(msg); */
        this.childrenToAdd = 1; // reinit to avoid confusion
    }

    public onMoveUpClick() {
        this.selectedItem.parent.moveChildUp(this.selectedItem);
        this.refresh();
    }

    public onMoveDownClick() {
        this.selectedItem.parent.moveChildDown(this.selectedItem);
        this.refresh();
    }

    public onRemoveClick() {
        this.selectedItem.parent.deleteChild(this.selectedItem.findPositionInParent());
        this.refresh();
    }

    // show modal dialog for values edition
    public showEditPab() {
        if (this.selectedItems.length > 0) {

            // list variables eligible to modification
            const availableVariables: { label: string, value: string, occurrences: number, first: number, last: number }[] = [];
            for (const c of this.selectedItems) {
                for (const p of c.parameterIterator) { // deep one
                    if (
                        p.visible &&
                        ! availableVariables.map(av => av.value).includes(p.symbol)
                    ) {
                        availableVariables.push({
                            label: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, p.symbol),
                            value: p.symbol,
                            occurrences: 0,
                            first: undefined,
                            last: undefined
                        });
                    }
                }
            }
            // find their min/max values (2nd pass)
            for (const av of availableVariables) {
                for (const c of this.selectedItems) {
                    for (const p of c.parameterIterator) {
                        if (p.visible && p.symbol === av.value) {
                            av.occurrences ++;
                            if (av.first === undefined) {
                                av.first = p.singleValue;
                            }
                            av.last = p.singleValue;
                        }
                    }
                }
            }

            // sum up selected items
            const walls: ParallelStructure[] = [];
            const wallsDevices: Structure[] = [];
            const devices: Structure[] = [];
            let vertical = true; // @TODO vertical AND consecutive !
            let firstDevicePosition: number;
            // 1st pass
            for (const s of this.selectedItems) {
                if (s instanceof ParallelStructure) {
                    walls.push(s);
                    for (const c of s.structures) {
                        if (firstDevicePosition === undefined) {
                            firstDevicePosition = c.findPositionInParent();
                        } else {
                            vertical = (vertical && (c.findPositionInParent() === firstDevicePosition));
                        }
                        wallsDevices.push(c);
                    }
                }
            }
            // 2nd pass
            for (const c of this.selectedItems) {
                if (c instanceof Structure) {
                    if (! wallsDevices.includes(c)) {
                        if (firstDevicePosition === undefined) {
                            firstDevicePosition = c.findPositionInParent();
                        } else {
                            vertical = (vertical && (c.findPositionInParent() === firstDevicePosition));
                        }
                        devices.push(c);
                    }
                }
            }

            // open dialog
            const dialogRef = this.editPabDialog.open(
                DialogEditPabComponent,
                {
                    data: {
                        availableVariables: availableVariables,
                        selectedItemsAbstract: {
                            walls: walls.length,
                            wallsDevices: wallsDevices.length,
                            devices: devices.length
                        },
                        vertical: vertical // used to enable interpolation
                    },
                    disableClose: true
                }
            );

            // apply modifications
            dialogRef.afterClosed().subscribe(result => {
                if (result) {
                    /* console.log("Apply values in parent !!", result.action, result.variable, result.value,
                      result.delta, result.variableDetails); */
                    switch (result.action) {
                        case "set-value":
                            for (const s of this.selectedItems) {
                                for (const p of s.parameterIterator) { // deep
                                    if (p.symbol === result.variable) {
                                        p.singleValue = result.value;
                                    }
                                }
                            }
                            break;

                        case "delta":
                            for (const s of this.selectedItems) {
                                for (const p of s.parameterIterator) { // deep
                                    if (p.symbol === result.variable) {
                                        p.singleValue += result.delta;
                                    }
                                }
                            }
                            break;

                        case "interpolate":
                            if (result.variableDetails.occurrences > 1) {
                                const nDigits = this.appSetupService.displayDigits;
                                // build values list
                                const interpolatedValues: number[] = [];
                                const step = (
                                    (result.variableDetails.last - result.variableDetails.first)
                                    / (result.variableDetails.occurrences - 1)
                                );
                                interpolatedValues.push(result.variableDetails.first);
                                let currentValue: number = result.variableDetails.first;
                                for (let i = 0; i < result.variableDetails.occurrences - 2; i++) {
                                    currentValue += step;
                                    interpolatedValues.push(Number(currentValue.toFixed(nDigits)));
                                }
                                interpolatedValues.push(result.variableDetails.last);
                                // apply
                                let idx = 0;
                                for (const s of this.selectedItems) {
                                    for (const p of s.parameterIterator) { // deep
                                        if (p.symbol === result.variable) {
                                            p.singleValue = interpolatedValues[idx];
                                            idx ++;
                                        }
                                    }
                                }
                            } else {
                                throw new Error(
                                    `showEditPab() : cannot interpolate, too few occurrences (${result.variableDetails.occurrences})`
                                );
                            }
                            break;
                    }
                }
            });
        }
    }

    /* public ngAfterViewInit() {
        this.onFieldsetListChange();
        this._fieldsetComponents.changes.subscribe(_ => this.onFieldsetListChange());
    }

    public ngDoCheck() {
        this.updateValidity();
    } */

    /**
     * @TODO Calcul de la validité de la Pab
     */
    private updateValidity() {
        this._isValid = false;

        /* if (this._fieldsetComponents !== undefined) {
            this._isValid = this._fieldsetComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    fieldset,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && fieldset.isValid;
                }
                // valeur initiale
                , this._fieldsetComponents.length > 0);
        } */

        this.validChange.emit();
    }

    /**
    * réception d'un événement de changement de valeur d'un input
    */
    private onInputChange($event) {
        this.inputChange.emit($event);
    }

    /**
     * Renvoie l'événement au composant du dessus
     */
    public onTabPressed(event) {
        console.log("tab pressed dans le tablo !");
    }

    public get uitextEditPabTable() {
        return "Modifier les valeurs";
    }
}
